/*
 * Copyright (c) 2014, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <ctype.h>

#include <aversive.h>
#include <aversive/endian.h>
#include <aversive/wait.h>
#include <aversive/pgmspace.h>

#include <callout.h>
#include <rdline.h>
#include <xbee.h>
#include <xbee_rxtx.h>

#include "rc_proto.h"
#include "xbee_user.h"
#include "main.h"

#define XBEE_TIMEOUT_MS    1000
#define XBEE_POLL_TIMER_MS 5

static struct xbee_ctx xbee_ctx[XBEE_MAX_CHANNEL];

int xbee_cmdline_input_enabled = 1;

/* parameters */
int xbee_raw = 0;

static struct callout xbee_rx_poll_timer;

static void __hexdump(const void *buf, unsigned int len)
{
	unsigned int i, out, ofs;
	const unsigned char *data = buf;
#define LINE_LEN 80
	char line[LINE_LEN];	/* space needed 8+16*3+3+16 == 75 */

	ofs = 0;
	while (ofs < len) {
		/* format 1 line in the buffer, then use printk to print them */
		out = snprintf_P(line, LINE_LEN, PSTR("%08X"), ofs);
		for (i=0; ofs+i < len && i<16; i++)
			out += snprintf_P(line+out, LINE_LEN - out,
					  PSTR(" %02X"),
					  data[ofs+i]&0xff);
		for (;i<=16;i++)
			out += snprintf(line+out, LINE_LEN - out, "   ");
		for (i=0; ofs < len && i<16; i++, ofs++) {
			unsigned char c = data[ofs];
			if (!isascii(c) || !isprint(c))
				c = '.';
			out += snprintf_P(line+out,
					  LINE_LEN - out,
					  PSTR("%c"), c);
		}
		DEBUG(E_USER_XBEE, "%s", line);
	}
}

static void hexdump_msg(const char *title, const struct xbee_msg *msg)
{
	unsigned i;

	DEBUG(E_USER_XBEE, "dump %s", title);

	for (i = 0; i < msg->iovlen; i++) {
		DEBUG(E_USER_XBEE, "iovec %d at %p, len=%d", i,
			msg->iov[i].buf, msg->iov[i].len);
		__hexdump(msg->iov[i].buf, msg->iov[i].len);
	}
}

void hexdump(const char *title, const void *buf, unsigned int len)
{
	DEBUG(E_USER_XBEE, "dump %s at [%p], len=%d", title, buf, len);
	__hexdump(buf, len);
}

static int parse_xmit_status(struct xbee_ctx *ctx,
	struct xbee_xmit_status_hdr *frame, unsigned len)
{
	(void)len;

	if (ctx == NULL) {
		ERROR(E_USER_XBEE, "no context");
		return -1;
	}

	/* see if it matches a xmit query (atcmd_query must be \0) */
	if (ctx->atcmd_query[0] != '\0') {
		ERROR(E_USER_XBEE, "invalid response 2");
		return -1;
	}

	/* XXX use defines for these values */
	if (frame->delivery_status == 0x00)
		NOTICE(E_USER_XBEE, "Success");
	else if (frame->delivery_status == 0x01)
		WARNING(E_USER_XBEE, "MAC ACK Failure");
	else if (frame->delivery_status == 0x15)
		WARNING(E_USER_XBEE, "Invalid destination endpoint");
	else if (frame->delivery_status == 0x21)
		WARNING(E_USER_XBEE, "Network ACK Failure");
	else if (frame->delivery_status == 0x25)
		WARNING(E_USER_XBEE, "Route Not Found");

	return 0;
}

/* Write in a human readable format the content of an atcmd response frame. It
 * assumes the frame is valid .*/
void atresp_to_str(char *buf, unsigned buflen, const struct xbee_atresp_hdr *frame,
	unsigned len)
{
	union {
		uint8_t u8;
		uint16_t u16;
		uint32_t u32;
		int16_t s16;
	} __attribute__((packed)) *result;
	char atcmd_str[3];
	const struct xbee_atcmd *cmd_pgm;
	struct xbee_atcmd cmd;

	/* get AT command from frame */
	memcpy(atcmd_str, &frame->cmd, 2);
	atcmd_str[2] = '\0';


	/* see if it exists */
	cmd_pgm = xbee_atcmd_lookup_name(atcmd_str);
	if (cmd_pgm == NULL) {
		snprintf(buf, buflen, "<%s> (unknown cmd)", atcmd_str);
		return;
	}
	memcpy_P(&cmd, cmd_pgm, sizeof(cmd));
	len -= sizeof(*frame);

	/* dump frame */
	result = (void *)frame->data;

	if (cmd.flags & XBEE_ATCMD_F_PARAM_U8 && len == sizeof(uint8_t))
		snprintf(buf, buflen, "<%s> is 0x%x (%d)", atcmd_str,
			result->u8, result->u8);
	else if (cmd.flags & XBEE_ATCMD_F_PARAM_U16 && len == sizeof(uint16_t))
		snprintf(buf, buflen, "<%s> is 0x%x (%d)", atcmd_str,
			ntohs(result->u16), ntohs(result->u16));
	else if (cmd.flags & XBEE_ATCMD_F_PARAM_U32 && len == sizeof(uint32_t))
		snprintf(buf, buflen, "<%s> is 0x%"PRIx32" (%"PRIu32")",
			atcmd_str, ntohl(result->u32), ntohl(result->u32));
	else if (cmd.flags & XBEE_ATCMD_F_PARAM_S16 && len == sizeof(int16_t))
		snprintf(buf, buflen, "<%s> is %d",
			atcmd_str, ntohs(result->s16));
	else if (len == 0)
		snprintf(buf, buflen, "<%s> no data", atcmd_str);
	else
		snprintf(buf, buflen, "invalid atresp");
}

static int parse_atcmd(struct xbee_ctx *ctx, struct xbee_atresp_hdr *frame,
	unsigned len)
{
	char atcmd_str[3];
	char buf[32];

	if (ctx == NULL) {
		ERROR(E_USER_XBEE, "no context");
		return -1;
	}

	/* get AT command from frame */
	memcpy(atcmd_str, &frame->cmd, 2);
	atcmd_str[2] = '\0';

	/* see if it matches query */
	if (memcmp(atcmd_str, ctx->atcmd_query, 2)) {
		ERROR(E_USER_XBEE, "invalid response <%c%c><%s><%s>",
			 frame->cmd & 0xFF,
			 (frame->cmd >> 8) & 0xFF,
			 atcmd_str, ctx->atcmd_query);
		return -1;
	}

	/* bad status */
	if (frame->status == 1) {
		WARNING(E_USER_XBEE, "Status is error");
		return -1;
	}
	else if (frame->status == 2) {
		WARNING(E_USER_XBEE, "Invalid command");
		return -1;
	}
	else if (frame->status == 3) {
		WARNING(E_USER_XBEE, "Invalid parameter");
		return -1;
	}
	else if (frame->status != 0) {
		WARNING(E_USER_XBEE, "Unknown status error %d",
			frame->status);
		return -1;
	}

	atresp_to_str(buf, sizeof(buf), frame, len);

	len -= sizeof(*frame);
	NOTICE(E_USER_XBEE, "status ok, datalen=%d, %s", len, buf);

	if (len != 0)
		hexdump("atcmd answer", frame->data, len);

	return 0;
}


/* Main xbee rx entry point for application. It decodes the xbee frame type and
 * dispatch to the application layer. Then "len" argument does not include the
 * xbee_hdr structure (delimiter, len, type, id) and checksum. */
int8_t xbeeapp_rx(struct xbee_dev *dev, int channel, int type,
	     void *frame, unsigned len, void *opaque)
{
	struct xbee_ctx *ctx = opaque;
	int8_t ret = XBEE_USER_RETCODE_OK;

	NOTICE(E_USER_XBEE, "type=0x%x, channel=%d, ctx=%p",
		type, channel, ctx);
	__hexdump(frame, len);

	/* if ctx is !NULL, it is an answer to a query */
	if (ctx != NULL) {
		xbee_unload_timeout(ctx);
		if (ctx->atcmd_query[0])
			NOTICE(E_USER_XBEE, "Received answer to query <%c%c>",
				ctx->atcmd_query[0], ctx->atcmd_query[1]);
	}

	/* some additional checks before sending */
	switch (type) {
		case XBEE_TYPE_MODEM_STATUS: {
			NOTICE(E_USER_XBEE, "Received Modem Status frame");
			break;
		}

		case XBEE_TYPE_RMT_ATRESP: {
			union {
				uint64_t u64;
				struct {
#if BYTE_ORDER == LITTLE_ENDIAN
					uint32_t low;
					uint32_t high;
#else
					uint32_t high;
					uint32_t low;
#endif
				} u32;
			} addr;
			memcpy(&addr, frame, sizeof(addr));
			addr.u64 = ntohll(addr.u64);
			NOTICE(E_USER_XBEE,
				"from remote address %"PRIx32"%"PRIx32"",
				 addr.u32.high, addr.u32.low);

			/* this answer contains an atcmd answer at offset 10 */
			if (parse_atcmd(ctx, frame + 10, len - 10) < 0)
				ret = XBEE_USER_RETCODE_BAD_FRAME;

			break;
		}
		case XBEE_TYPE_ATRESP: {
			if (parse_atcmd(ctx, frame, len) < 0)
				ret = XBEE_USER_RETCODE_BAD_FRAME;

			break;
		}

		case XBEE_TYPE_XMIT_STATUS: {
			if (parse_xmit_status(ctx, frame, len) < 0)
				ret = XBEE_USER_RETCODE_BAD_FRAME;

			break;
		}

		case XBEE_TYPE_RECV: {
			if (rc_proto_rx(frame, len) < 0)
				ret = XBEE_USER_RETCODE_BAD_FRAME;

			break;
		}

		case XBEE_TYPE_ATCMD:
		case XBEE_TYPE_ATCMD_Q:
		case XBEE_TYPE_XMIT:
		case XBEE_TYPE_EXPL_XMIT:
		case XBEE_TYPE_RMT_ATCMD:
		case XBEE_TYPE_EXPL_RECV:
		case XBEE_TYPE_NODE_ID:
		default:
			ERROR(E_USER_XBEE, "Invalid frame");
			ret = XBEE_USER_RETCODE_BAD_FRAME;
			break;
	}

	if (ret != XBEE_USER_RETCODE_OK) {
		WARNING(E_USER_XBEE, "undecoded rx frame");
	}

	if (ctx != NULL) {
		/* callback */
		if (ctx->rx_cb != NULL)
			ret = ctx->rx_cb(ret, frame, len, ctx->arg);

		/* free channel now, it implicitely frees the context too */
		xbee_unregister_channel(dev, channel);
	}

	return ret;
}

/* called with callout prio == XBEE_PRIO */
static int xbeeapp_send(struct xbee_ctx *ctx, int type, struct xbee_msg *msg)
{
	int ret;
	int channel;

	/* register a channel */
	channel = xbee_register_channel(xbee_dev, XBEE_CHANNEL_ANY,
					xbeeapp_rx, NULL);
	if (channel < 0) {
		ERROR(E_USER_XBEE, "cannot send: no free channel");
		return -1;
	}

	/* copy context in the static struct table (avoiding a malloc) */
	memcpy(&xbee_ctx[channel], ctx, sizeof(*ctx));
	ctx = &xbee_ctx[channel];
	xbee_set_opaque(xbee_dev, channel, ctx);

	NOTICE(E_USER_XBEE, "send frame channel=%d type=0x%x", channel, type);
	hexdump_msg("xmit frame", msg);

	/* transmit the frame on this channel */
	ret = xbee_tx_iovec(xbee_dev, channel, type, msg);
	if (ret < 0) {
		ERROR(E_USER_XBEE, "cannot send");
		xbee_unregister_channel(xbee_dev, channel);
		return -1;
	}

	ctx->channel = channel;
	xbee_load_timeout(ctx);   /* load a timeout event */

	return 0;
}

/* send an AT command with parameters filled by caller. param is the argument
 * of the atcmd (if any). */
int xbeeapp_send_atcmd(char *atcmd_str, void *param, unsigned param_len,
	xbee_user_rx_cb_t *rx_cb, void *arg)
{
	struct xbee_ctx ctx;
	/* struct xbee_atcmd_hdr atcmd_hdr; -> no needed same than atcmd_str */
	struct xbee_msg msg;
	uint8_t prio;
	int ret;

	memset(&ctx, 0, sizeof(ctx));
	ctx.atcmd_query[0] = atcmd_str[0];
	ctx.atcmd_query[1] = atcmd_str[1];
	ctx.rx_cb = rx_cb;
	ctx.arg = arg;

	if (param_len == 0) {
		msg.iovlen = 1;
		msg.iov[0].buf = atcmd_str;
		msg.iov[0].len = 2;
	}
	else {
		msg.iovlen = 2;
		msg.iov[0].buf = atcmd_str;
		msg.iov[0].len = 2;
		msg.iov[1].buf = param;
		msg.iov[1].len = param_len;
	}

	prio = callout_mgr_set_prio(&xbeeboard.intr_cm, XBEE_PRIO);
	if (prio > XBEE_PRIO)
		ERROR(E_USER_XBEE, "invalid prio = %d\n", prio);
	ret = xbeeapp_send(&ctx, XBEE_TYPE_ATCMD, &msg);
	callout_mgr_restore_prio(&xbeeboard.intr_cm, prio);

	return ret;
}

int xbeeapp_send_msg(uint64_t addr, struct xbee_msg *msg,
	xbee_user_rx_cb_t *rx_cb, void *arg)
{
	struct xbee_ctx ctx;
	struct xbee_xmit_hdr xmit_hdr;
	struct xbee_msg msg2;
	unsigned i;
	uint8_t prio;
	int ret;

	if (msg->iovlen + 2 > XBEE_MSG_MAXIOV) {
		ERROR(E_USER_XBEE, "too many iovecs");
		return -1;
	}

	xmit_hdr.dstaddr = htonll(addr);
	xmit_hdr.reserved = htons(0xFFFE);
	xmit_hdr.bcast_radius = 0;
	xmit_hdr.opts = 0;

	msg2.iovlen = msg->iovlen + 1;
	msg2.iov[0].buf = &xmit_hdr;
	msg2.iov[0].len = sizeof(xmit_hdr);
	for (i = 0; i < msg->iovlen; i++)
		msg2.iov[i+1] = msg->iov[i];

	memset(&ctx, 0, sizeof(ctx));
	ctx.rx_cb = rx_cb;
	ctx.arg = arg;

	prio = callout_mgr_set_prio(&xbeeboard.intr_cm, XBEE_PRIO);
	if (prio > XBEE_PRIO)
		ERROR(E_USER_XBEE, "invalid prio = %d\n", prio);
	ret = xbeeapp_send(&ctx, XBEE_TYPE_XMIT, &msg2);
	callout_mgr_restore_prio(&xbeeboard.intr_cm, prio);

	return ret;
}

static void evt_timeout(struct callout_mgr *cm, struct callout *clt,
	void *arg)
{
	struct xbee_ctx *ctx = arg;

	(void)cm;
	(void)clt;

	WARNING(E_USER_XBEE, "Timeout");

	/* callback */
	if (ctx->rx_cb != NULL)
		ctx->rx_cb(XBEE_USER_RETCODE_TIMEOUT, NULL, 0, ctx->arg);

	/* free channel, it implicitely frees the context too */
	xbee_unregister_channel(xbee_dev, ctx->channel);
	callout_stop(cm, clt);
}

void xbee_load_timeout(struct xbee_ctx *ctx)
{
	uint8_t prio;

	callout_init(&ctx->timeout, evt_timeout, ctx, XBEE_PRIO);
	prio = callout_mgr_set_prio(&xbeeboard.intr_cm, XBEE_PRIO);
	callout_schedule(&xbeeboard.intr_cm, &ctx->timeout, XBEE_TIMEOUT_MS);
	callout_mgr_restore_prio(&xbeeboard.intr_cm, prio);
}

void xbee_unload_timeout(struct xbee_ctx *ctx)
{
	uint8_t prio;

	prio = callout_mgr_set_prio(&xbeeboard.intr_cm, XBEE_PRIO);
	callout_stop(&xbeeboard.intr_cm, &ctx->timeout);
	callout_mgr_restore_prio(&xbeeboard.intr_cm, prio);
}

static void xbee_rx_poll_timer_cb(struct callout_mgr *cm, struct callout *tim,
	void *arg)
{
	(void) arg;
	xbee_rx(xbee_dev);
	callout_reschedule(cm, tim, XBEE_POLL_TIMER_MS);
}

void xbee_mainloop(void)
{
	uint8_t prio;

	while (1) {
		if (xbee_raw) {
			int16_t c;

			/* from xbee to cmdline */
			c = xbee_dev_recv(NULL);
			if (c >= 0)
				cmdline_dev_send((uint8_t)c, NULL);

			/* from cmdline to xbee */
			c = cmdline_dev_recv(NULL);
			if (c == 4) { /* CTRL-d */
				prio = callout_mgr_set_prio(&xbeeboard.intr_cm,
					XBEE_PRIO);
				xbee_dev_send('A', NULL);
				xbee_dev_send('T', NULL);
				xbee_dev_send('C', NULL);
				xbee_dev_send('N', NULL);
				xbee_dev_send('\n', NULL);
				callout_mgr_restore_prio(&xbeeboard.intr_cm, prio);
				xbee_raw = 0;
				rdline_newline(&xbeeboard.rdl,
					       xbeeboard.prompt);
			}
			else if (c >= 0) {
				/* send to xbee */
				xbee_dev_send((uint8_t)c, NULL);

				/* echo on cmdline */
				cmdline_dev_send((uint8_t)c, NULL);
			}
		}
		else {
			if (xbee_cmdline_input_enabled)
				cmdline_poll();
			/* xbee rx polling is done in a timer, so we can block
			 * the cmdline without loosing incoming packets */
		}
	}
}

void xbee_stdin_enable(void)
{
	xbee_cmdline_input_enabled = 1;
}

void xbee_stdin_disable(void)
{
	xbee_cmdline_input_enabled = 0;
}

void xbeeapp_init(void)
{
	callout_init(&xbee_rx_poll_timer, xbee_rx_poll_timer_cb,
		NULL, XBEE_PRIO);
	callout_schedule(&xbeeboard.intr_cm, &xbee_rx_poll_timer,
		XBEE_POLL_TIMER_MS);
}
